/* 
 * UC:21111 - Sistemas Operativos
 * e-fólio A 2020-21
 *
 * Aluno: 2000027 - Hélio Sousa
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dirent.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>

/**Nas funções que não têm envp como argumento, as variáveis de ambiente 
 * são as definidas na variável global environ.
 */
extern char **environ;

/**
 * Teste ao numero de argumentos, caso este seja diferente de 2 informa o utilizador
 * e sai do programa.(dois argumentos pois o nome do comando conta como um argumento)
 * atenção que argc conta também o comando em si, pelo que temos de contabilizar 2
 * um para o comando e outro para o diretório passado
 * @param argc numero de argumentos passados á função main()
 * @return 0 sucesso na validação e -1 insucesso.
 */
int valida_args ( int argc ) { 				

    if (argc != 2) {			
        printf ("ERRO: número de argumentos inválido.\n");  
        return -1;     
    }
    else {
        return 0;      
    }
}

/**
 * Optei pela função opendir() em vez do fopen, porque na fase de testes fopen 
 * retornava Ok para um ficheiro e apenas pretendemos dirirotios. Por isso, procurei
 * uma função que indicasse se um dado diretorio existe ou não, mas integrado
 * na norma UNIX e standard C, e opendir() cumpre esses critérios.
 * @param argv diretorio a ser introduzido
 * @return 0 sucesso na validação e -1 insucesso.
 */
int valida_dir (char *argv[]) { 				

    DIR *dirptr = opendir(argv[1]);

    if (ENOENT == errno) {			
        printf ("Diretório inválido ou inexistente: %s\n",argv[1]);
        return -1;
    }
    else {        
        closedir(dirptr);
        return 0; 
    }
}

/**
 * Verifica a criação de um novo processo
 * @param pid numero do processo do pai
 * @return 0 sucesso na validação e -1 insucesso.
 */
int fork_check(pid_t pid) {
    if ( pid < 0 ) {
        printf ("Erro na funcao fork()");
        return -1;
    }
    else {
        return 0;
    }     
}

/**
 * Imprime para a tela os dados de um processo.
 * A denominação (A,B,...), o PID e o PPID.
 * @param ch letra do processo
 */
void msg_processo(char ch) {
   printf("Processo %c: PID=%5d  PPID=%5d\n", ch, (int) getpid(), (int) getppid());
}

/**
 * retorna mensagem erro quando ocorre erro na execução de um programa
 * exec??
 * @param ch letra que identifica o processo
 */
void msg_erro_processo (char ch) {
    printf("Se esta mensagem aparecer ocorreu um erro no processo%c!",ch);    
}

/**
 * Primeiro redireciona-se o stdout para o ficheiro tmp1.txt.
 * Depois executa-se a troca da imagem com o execl
 * @param argv diretorio passado na linha de comandos.
 */
void processo_Bls(char *argv[]) {     
    freopen("tmp1.txt", "w+", stdout);
    execl("/bin/ls","ls","-1",argv[1], NULL);
    msg_erro_processo('B');
}  

/**
 * Acrescenta no ficheiro tmp1.txt o texto "Itens encontrados: "
 * @param argv diretorio passado na linha de comandos.
 */
void processo_Cecho() {
    freopen("tmp1.txt", "a+", stdout);
    execlp("echo","echo","-n", "Itens encontrados: ", NULL);
    msg_erro_processo('C'); 
}

/**
 * Redireciona o stout para o ficheiro tmp2.txt.
 * Redireciona o stdin para o ficheiro tmp1.txt
 * O comando wc -l lê o standard input se nenhum ficheiro for passado
 * na linha de comandos.
 * O seu output é redirecionado para o tmp2.txt
 */
void processo_Dwc() {
    freopen("tmp1.txt", "r", stdin);  
    freopen("tmp2.txt", "w+", stdout); 
    char *arguments[1];
    arguments[0] = "wc";
    arguments[1] = "-l";            
    arguments[2] = NULL;
    execv( "/bin/wc", arguments);
    msg_erro_processo('D');
}

/**
 * Executa a concatenação dos ficheiros tmp1.txt e tmp2.txt
 * e imprime o resultado final para o ecrã.
 * Utiliza a função execvp que recebe como parametro o vetor arguments.
 */
void processo_Ecat() {
    char *arguments[3];
    arguments[0] = "cat";
    arguments[1] = "tmp1.txt";            
    arguments[2] = "tmp2.txt";
    arguments[3] = NULL;    
    execvp("cat", arguments); 
    msg_erro_processo('E');
}


/**
 * Programa dit ou Processo A. Será a partir deste programa que serão derivados
 * 4 processos sequenciais gerados pelo mesmo pai (Processo A).
 * @param argc contabiliza o numero de argumentos, neste caso, apenas se admite 2, 
 * o próprio programa ./dit e o diretório a ser analisado
 * @param argv retorna um apontador para o diretorio a ser analisado
 * @return retorna uma lista do conteudo da diretoria contendo um item por linha,
 * e adiciona no final o numero de itens da lista e imprime a lista na saída
 * padrão (stout).
 */
int main(int argc, char** argv) {

    pid_t pidB,pidC,pidD,pidE;
    int status; 
       
    // Processo A 
    if (valida_args(argc)) return (EXIT_FAILURE); 
    if (valida_dir(argv)) return (EXIT_FAILURE);     
    
    pidB=fork();
    if (fork_check(pidB)) return (EXIT_FAILURE);
        
    if ( pidB == 0 ) {
        // Processo B
        msg_processo('B');
        processo_Bls(argv);                  
    }
    else {
        msg_processo('A');
        waitpid(pidB,&status,0);
        pidC=fork();
        if (fork_check(pidC)) return (EXIT_FAILURE);
        if (pidC == 0) {
            // Processo C
            msg_processo('C');
            processo_Cecho();
        }
        else {
            waitpid(pidC,&status,0);
            pidD=fork();
            if (fork_check(pidD)) return (EXIT_FAILURE);           
            if (pidD == 0){
                // Processo D
                msg_processo('D');              
                processo_Dwc();
            }
            else {
                waitpid(pidD,&status,0);
                pidE=fork();
                if (fork_check(pidE)) return (EXIT_FAILURE);                  
                    if (pidE == 0) {
                        // Processo E
                        msg_processo('E');                 
                        processo_Ecat();
                    }
                    else {
                        waitpid(pidE,&status,0);                                            
                        //termina processoA(dit);
                    }
            }         
        }                  
    }  
    return (EXIT_SUCCESS);
}
